// Copyright 2008 nod_chip
// Copyright 2008 Aroo		(Vectorization with SSE2)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "stdafx.h"
#include <malloc.h>

#include "ProcessorSse2Aroo.h"

// stdafx.h_USE_MATH_DEFINES`ĂȂ̂
#define M_LN2      0.693147180559945309417

namespace {

// wikipediaen:exponential_functionQl
// [-88.?? ,88.??]ɍiăxNg
// exp(x),x<-870ĂƂ
// u : [0,ln(2)]Ȃ̂łƂ肠2őł؂ <- xቺ̌

static const __m128 ln2 = _mm_set_ps1(M_LN2);
static const __m128 inv_ln2 = _mm_set_ps1(1/M_LN2);
static const __m128 f1 = _mm_set_ps1(1);
static const __m128 inv_f2 = _mm_set_ps1(0.5);
static const __m128i exp_nmin = _mm_set1_epi32(-128);
static const __m128i exp_bias = _mm_set1_epi32(127);

inline __m128 exp(const __m128 x)
{
	const __m128i n = _mm_cvttps_epi32(_mm_mul_ps(x,inv_ln2));
	const __m128 u = _mm_sub_ps(x,_mm_mul_ps(_mm_cvtepi32_ps(n),ln2));
	const __m128 m = _mm_add_ps(_mm_mul_ps(_mm_mul_ps(u,u),inv_f2),_mm_add_ps(u,f1));
	const __m128i tPofn = _mm_slli_epi32(_mm_add_epi32(n,exp_bias),23);
	const __m128i mask = _mm_cmpgt_epi32(n,exp_nmin);
	return _mm_and_ps(_mm_mul_ps(m,*(const __m128*)&tPofn),*(const __m128*)&mask);
}

// ȉQlɒPxł쐬
// x悤Ȃ
// http://citeseer.ist.psu.edu/schraudolph98fast.html
// y[31:16] = a * x + b - c
// a = (2^(23-16))/ln(2)
// 23 : frac bit width
// 16 : integer(short) bit width
// b = 127 * 2^(23-16)
// c = ?

static const __m128 fastexp_a = _mm_set_ps1(0x80/M_LN2);
static const __m128i fastexp_bmc = _mm_set1_epi32(127*0x80 - 9);
static const __m128 fastexp_min = _mm_set_ps1(-87);

inline __m128 fastexp(const __m128 x)
{
	union {
		__m128 f;
		__m128i i;
	};
	const __m128 mask = _mm_cmpge_ps(x,fastexp_min);
	const __m128i ax = _mm_cvtps_epi32(_mm_mul_ps(x,fastexp_a));
	i = _mm_and_si128(_mm_slli_epi32(_mm_add_epi32(ax,fastexp_bmc),16),*(__m128i*)&mask);
	return f;
}

// ꉞ
inline __m128 refexp(__m128 x)
{
	_mm_empty();
	for(int i=0;i<4;++i) {
		x.m128_f32[i] = std::exp(x.m128_f32[i]);
	}
	return x;
}

void copyS3toS4(__m128i * const pOut,
				const int pitch_m128,
				const PIXEL_YC * const pIn,
				const int w,
				const int h,
				const int max_w,
				const int clamp_w)
{
	const int pitch = pitch_m128;
	const int xcount = (w+7)/8;
	const int sp = w%8;
	const PIXEL_YC *pCrntIn = pIn;
	__m128i *pCrntOut = pOut;

	__m64 head = *(__m64*)pCrntIn;
	__m64 *pCrntOutClamp = (__m64*)pCrntOut - clamp_w;
	for( int x=0; x < clamp_w; ++x) {
		_mm_stream_pi(pCrntOutClamp++,head);
	}

	for( int x=0; x < xcount; ++x ) {
		const __m128i a = _mm_load_si128((__m128i*)pCrntIn);
		const __m128i b = _mm_load_si128((__m128i*)pCrntIn + 1);
		const __m128i c = _mm_load_si128((__m128i*)pCrntIn + 2);
		__m128i k = _mm_slli_si128(a,2);
		k = _mm_shufflelo_epi16(k,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut, k);
		__m128i l = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&a,*(__m128d*)&b,_MM_SHUFFLE2(0,1));
		l = _mm_srli_si128(l,2);
		l = _mm_shufflelo_epi16(l,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+1, l);
		__m128i m = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&b,*(__m128d*)&c,_MM_SHUFFLE2(0,1));
		m = _mm_slli_si128(m,2);
		m = _mm_shufflelo_epi16(m,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+2, m);
		__m128i n = _mm_srli_si128(c,2);
		n = _mm_shufflelo_epi16(n,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+3, n);
		__m128i *pCrntOutClamp = pCrntOut;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp-=pitch,k);
		}
		pCrntOutClamp = pCrntOut + 1;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp-=pitch,l);
		}
		pCrntOutClamp = pCrntOut + 2;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp-=pitch,m);
		}
		pCrntOutClamp = pCrntOut + 3;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp-=pitch,n);
		}
		pCrntIn += 8;
		pCrntOut += 4;
	}

	__m64 tail;
	tail = *(const __m64*)(pCrntIn - sp - 1);
	pCrntOutClamp = ((__m64*)pCrntOut - sp - 1);
	for(int y=0;y<clamp_w;++y) {
		_mm_stream_pi(pCrntOutClamp-=2*pitch,tail);
	}
	pCrntOutClamp = ((__m64*)pCrntOut - sp);
	for( int x=0; x < clamp_w; ++x) {
		_mm_stream_pi(pCrntOutClamp++,tail);
	}

	int y;
//#pragma omp parallel for
	for( y = 1; y < h-1; ++y ) {
		const PIXEL_YC *pCrntInX = pIn + max_w * y;
		__m128i *pCrntOutX = pOut + pitch * y;

		__m64 *pCrntOutXClamp = (__m64*)pCrntOutX - clamp_w;
		head = *(__m64*)pCrntInX;
		for( int x=0; x < clamp_w; ++x) {
			_mm_stream_pi(pCrntOutXClamp++,head);
		}

		for( int x=0; x < xcount; ++x ) {
			const __m128i a = _mm_load_si128((__m128i*)pCrntInX);
			const __m128i b = _mm_load_si128((__m128i*)pCrntInX + 1);
			const __m128i c = _mm_load_si128((__m128i*)pCrntInX + 2);
			__m128i k = _mm_slli_si128(a,2);
			k = _mm_shufflelo_epi16(k,_MM_SHUFFLE(0,3,2,1));
			_mm_stream_si128(pCrntOutX++, k);
			__m128i l = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&a,*(__m128d*)&b,_MM_SHUFFLE2(0,1));
			l = _mm_srli_si128(l,2);
			l = _mm_shufflelo_epi16(l,_MM_SHUFFLE(0,3,2,1));
			_mm_stream_si128(pCrntOutX++, l);
			__m128i m = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&b,*(__m128d*)&c,_MM_SHUFFLE2(0,1));
			m = _mm_slli_si128(m,2);
			m = _mm_shufflelo_epi16(m,_MM_SHUFFLE(0,3,2,1));
			_mm_stream_si128(pCrntOutX++, m);
			__m128i n = _mm_srli_si128(c,2);
			n = _mm_shufflelo_epi16(n,_MM_SHUFFLE(0,3,2,1));
			_mm_stream_si128(pCrntOutX++, n);
			pCrntInX += 8;
		}

		tail = *(const __m64*)(pCrntInX - sp - 1);
		pCrntOutXClamp = ((__m64*)pCrntOutX - sp);
		for( int x=0; x < clamp_w; ++x) {
			_mm_stream_pi(pCrntOutXClamp++,tail);
		}
	}

	pCrntIn = pIn + max_w * (h-1);
	pCrntOut = pOut + pitch * (h-1);

	head = *(__m64*)pCrntIn;
	pCrntOutClamp = (__m64*)pCrntOut - clamp_w;
	for( int x=0; x < clamp_w; ++x) {
		_mm_stream_pi(pCrntOutClamp++,head);
	}
	for( int x=0; x < xcount; ++x ) {
		const __m128i a = _mm_load_si128((__m128i*)pCrntIn);
		const __m128i b = _mm_load_si128((__m128i*)pCrntIn + 1);
		const __m128i c = _mm_load_si128((__m128i*)pCrntIn + 2);
		__m128i k = _mm_slli_si128(a,2);
		k = _mm_shufflelo_epi16(k,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut, k);
		__m128i l = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&a,*(__m128d*)&b,_MM_SHUFFLE2(0,1));
		l = _mm_srli_si128(l,2);
		l = _mm_shufflelo_epi16(l,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+1, l);
		__m128i m = *(__m128i*)&_mm_shuffle_pd(*(__m128d*)&b,*(__m128d*)&c,_MM_SHUFFLE2(0,1));
		m = _mm_slli_si128(m,2);
		m = _mm_shufflelo_epi16(m,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+2, m);
		__m128i n = _mm_srli_si128(c,2);
		n = _mm_shufflelo_epi16(n,_MM_SHUFFLE(0,3,2,1));
		_mm_stream_si128(pCrntOut+3, n);
		__m128i *pCrntOutClamp = pCrntOut;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp+=pitch,k);
		}
		pCrntOutClamp = pCrntOut + 1;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp+=pitch,l);
		}
		pCrntOutClamp = pCrntOut + 2;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp+=pitch,m);
		}
		pCrntOutClamp = pCrntOut + 3;
		for(int y=0;y<clamp_w;++y) {
			_mm_stream_si128(pCrntOutClamp+=pitch,n);
		}
		pCrntIn += 8;
		pCrntOut += 4;
	}
	tail = *(const __m64*)(pCrntIn - sp - 1);
	pCrntOutClamp = ((__m64*)pCrntOut - sp - 1);
	for(int y=0;y<clamp_w;++y) {
		_mm_stream_pi(pCrntOutClamp+=2*pitch,tail);
	}
	pCrntOutClamp = ((__m64*)pCrntOut - sp);
	for( int x=0; x < clamp_w; ++x) {
		_mm_stream_pi(pCrntOutClamp++,tail);
	}

	__m64 tl = *(__m64*)&pIn[0];
	__m64 tr = *(__m64*)&pIn[w-1];
	__m64 bl = *(__m64*)&pIn[max_w * (h-1)];
	__m64 br = *(__m64*)&pIn[max_w * (h-1) + w-1];
	__m64 *ptl = (__m64*)pOut - (2*pitch+1) * clamp_w;
	__m64 *ptr = ptl + clamp_w + w;
	__m64 *pbl = ptl + 2 * pitch * (clamp_w + h);
	__m64 *pbr = pbl + clamp_w + w;
	for( int y = 0; y < clamp_w; ++y) {
		for ( int x = 0; x < clamp_w; ++x) {
			_mm_stream_pi(ptl + pitch * 2 * y + x, tl);
			_mm_stream_pi(ptr + pitch * 2 * y + x, tr);
			_mm_stream_pi(pbl + pitch * 2 * y + x, bl);
			_mm_stream_pi(pbr + pitch * 2 * y + x, br);
		}
	}
}

};

extern int track_e[];

ProcessorSse2Aroo::AllocMemory::AllocMemory()
: pitch(0),w(0),h(0),s(0),c(0),ap(NULL),max_radius(((track_e[0] + 1) / 2) * 2) // + 1 neighborhoodRadius
{}

ProcessorSse2Aroo::AllocMemory::~AllocMemory()
{
	_aligned_free(ap);
}

__m64* ProcessorSse2Aroo::AllocMemory::allocate(int nw, int nh, int nc)
{
	if(nw == w && nh == h && nc == c)
		return ap;

	int np = nw + 2 * max_radius;
	int ns = np * (nh + 2 * max_radius);
	if(ns!=s || nc != c) {
		deallocate();
		ap = (__m64*)_aligned_malloc(ns * sizeof(__m64) * nc,16);
		if(ap) {
			for(int i=0;i<nc;++i) {
				pv[i] = ap + ns * i + (np + 1) * max_radius;
			}
			pitch = np;
			w = nw;
			h = nh;
			s = ns;
			c = nc;
		}
	} else {
		w = nw;
		h = nh;
		pitch = np;
	}
	return ap;
}

void ProcessorSse2Aroo::AllocMemory::deallocate()
{
	_aligned_free(ap);
	pitch = w = h = s = c = 0;
	std::fill_n(pv,_countof(pv),(__m64*)NULL);
	ap = NULL;
}

ProcessorSse2Aroo::ProcessorSse2Aroo()
: prepared(IsProcessorFeaturePresent(PF_XMMI64_INSTRUCTIONS_AVAILABLE))
, currentYcpFilteringCacheSize(-1)
{
}

ProcessorSse2Aroo::~ProcessorSse2Aroo()
{
}

BOOL ProcessorSse2Aroo::proc(FILTER& fp, FILTER_PROC_INFO& fpip)
{
	//	fp->track[n]			: gbNo[̐l
	//	fp->check[n]			: `FbN{bNX̐l
	//	fpip->w 				: ۂ̉摜̉
	//	fpip->h 				: ۂ̉摜̏c
	//	fpip->max_w				: 摜̈̉
	//	fpip->max_h				: 摜̈̏c
	//	fpip->ycp_edit			: 摜̈ւ̃|C^
	//	fpip->ycp_temp			: e|̈ւ̃|C^
	//	fpip->ycp_edit[n].y		: f(Px    )f[^ (     0 ` 4095 )
	//	fpip->ycp_edit[n].cb	: f(F())f[^ ( -2048 ` 2047 )
	//	fpip->ycp_edit[n].cr	: f(F())f[^ ( -2048 ` 2047 )

	const int w = fpip.w;
	const int h = fpip.h;
	const int max_w = fpip.max_w;
	const int max_h = fpip.max_h;

	const int searchRadius = fp.track[0];
	const int neighborhoodRadius = 1;
	const float H = pow(10, (float)fp.track[2] / 22.0f);
	const __m128 mH2 = _mm_set_ps1(-1.0f / (H * H));

	const int timeSearchRadius = fp.track[1];
	const int timeSearchDiameter = timeSearchRadius * 2 + 1;

	const int clamp_w = searchRadius+neighborhoodRadius;
	const int pitch = mem.pitch;

	if( !mem.allocate(max_w,max_h,timeSearchDiameter) ) {
		return FALSE;
	}

	//LbVTCYw肷
	if (currentYcpFilteringCacheSize != timeSearchDiameter){
		fp.exfunc->set_ycp_filtering_cache_size(&fp, w, h, timeSearchDiameter, NULL);
		currentYcpFilteringCacheSize = timeSearchDiameter;
	}

	//t[ǂݍ
	const int currentFrame = fp.exfunc->get_frame(fpip.editp);
	const int numberOfFrames = fp.exfunc->get_frame_n(fpip.editp);
	for (int dt = -timeSearchRadius; dt <= timeSearchRadius; ++dt){
		const int targetFrame = max(0, min(numberOfFrames - 1, currentFrame + dt));
		const __m64 *p = mem.pv[timeSearchRadius + dt];
		PIXEL_YC *q = fp.exfunc->get_ycp_filtering_cache_ex(&fp, fpip.editp, targetFrame, NULL, NULL);
		copyS3toS4((__m128i*)p, pitch/2, q, w, h, w, clamp_w);
	}

	__m64 ** p = mem.pv;
	PIXEL_YC *q = fpip.ycp_edit;
	const int xcount = (w+1)/2;

	int y;
#pragma omp parallel for
	for(y = 0; y < h; ++y) {
		for(int x = 0; x < xcount; ++x) {
			__m128 usum = _mm_setzero_ps();
			__m128 uvalue = _mm_setzero_ps();
			__m128 vsum = _mm_setzero_ps();
			__m128 vvalue = _mm_setzero_ps();
			for(int dy = -searchRadius; dy <= searchRadius; ++dy) {
				for(int dx = -searchRadius; dx <= searchRadius; ++dx) {
					__m128 usum2 = _mm_setzero_ps();
					__m128 vsum2 = _mm_setzero_ps();
					for (int dt = -timeSearchRadius; dt <= timeSearchRadius; ++dt){
						const __m64 *mp = p[timeSearchRadius];
						const __m64 *np = p[timeSearchRadius + dt];
						for(int dy2 = -neighborhoodRadius; dy2 <= neighborhoodRadius; ++dy2) {
							__m128i u1 = _mm_loadu_si128((const __m128i*)(mp + pitch * (y      + dy2) + 2 * x      - 1));
							__m128i u2 = _mm_load_si128 ((const __m128i*)(mp + pitch * (y      + dy2) + 2 * x         ));
							__m128i u3 = _mm_loadu_si128((const __m128i*)(mp + pitch * (y      + dy2) + 2 * x      + 1));

							__m128i v1;
							__m128i v2;
							__m128i v3;

							if(dx&1) {
								v1 = _mm_load_si128 ((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx - 1));
								v2 = _mm_loadu_si128((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx    ));
								v3 = _mm_load_si128 ((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx + 1));
							} else {
								v1 = _mm_loadu_si128((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx - 1));
								v2 = _mm_load_si128 ((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx    ));
								v3 = _mm_loadu_si128((const __m128i*)(np + pitch * (y + dy + dy2) + 2 * x + dx + 1));
							}

							u1 = _mm_sub_epi16(u1,v1);
							u2 = _mm_sub_epi16(u2,v2);
							u3 = _mm_sub_epi16(u3,v3);

							v1 = _mm_mulhi_epi16(u1,u1);
							v2 = _mm_mulhi_epi16(u2,u2);
							v3 = _mm_mulhi_epi16(u3,u3);
							u1 = _mm_mullo_epi16(u1,u1);
							u2 = _mm_mullo_epi16(u2,u2);
							u3 = _mm_mullo_epi16(u3,u3);

							u1 = _mm_unpacklo_epi16(u1,v1);
							u2 = _mm_unpacklo_epi16(u2,v2);
							u3 = _mm_unpacklo_epi16(u3,v3);
							v1 = _mm_unpacklo_epi16(u1,v1);
							v2 = _mm_unpacklo_epi16(u2,v2);
							v3 = _mm_unpacklo_epi16(u3,v3);

							u1 = _mm_add_epi32(u1,u2);
							v1 = _mm_add_epi32(v1,v2);
							u1 = _mm_add_epi32(u1,u3);
							v1 = _mm_add_epi32(v1,v3);
							usum2 = _mm_add_ps(usum2,_mm_cvtepi32_ps(u1));
							vsum2 = _mm_add_ps(usum2,_mm_cvtepi32_ps(v1));
						}
						// x
						//usum2 = exp(_mm_mul_ps(usum2,mH2));
						//vsum2 = exp(_mm_mul_ps(vsum2,mH2));
						// x
						usum2 = fastexp(_mm_mul_ps(usum2,mH2));
						vsum2 = fastexp(_mm_mul_ps(vsum2,mH2));

						usum = _mm_add_ps(usum,usum2);
						vsum = _mm_add_ps(vsum,vsum2);

						__m128i v = _mm_loadu_si128((const __m128i*)(np + pitch * (y + dy) + 2 * x + dx));
						__m128i u = _mm_unpacklo_epi16(v,v);
						v = _mm_unpackhi_epi16(v,v);
						u = _mm_srai_epi32(u,16);
						v = _mm_srai_epi32(v,16);
						__m128 uf = _mm_cvtepi32_ps(u);
						__m128 vf = _mm_cvtepi32_ps(v);

						uf = _mm_mul_ps(uf,usum2);
						vf = _mm_mul_ps(vf,vsum2);
						uvalue = _mm_add_ps(uvalue,uf);
						vvalue = _mm_add_ps(vvalue,vf);
					}
				}
			}
			// x
			//uvalue = _mm_div_ps(uvalue,usum);
			//vvalue = _mm_div_ps(vvalue,vsum);
			// x
			// 11bitxt̐ςȂ̂ŐxႵ
			usum = _mm_rcp_ps(usum);
			vsum = _mm_rcp_ps(vsum);
			uvalue = _mm_mul_ps(uvalue,usum);
			vvalue = _mm_mul_ps(vvalue,vsum);

			__m128i valuei = _mm_cvtps_epi32(uvalue);
			__m128i vvaluei = _mm_cvtps_epi32(vvalue);
			valuei = _mm_packs_epi32(valuei,vvaluei);
			valuei = _mm_shufflelo_epi16(valuei,_MM_SHUFFLE(2,1,0,3));
			valuei = _mm_srli_si128(valuei,2);
			_mm_stream_pi((__m64*)&q[max_w * y + 2 * x],*(__m64*)&valuei);
			_mm_stream_si32((int*)&q[max_w * y + 2 * x] + 2,valuei.m128i_i32[2]);
		}
	}

	_mm_empty();

	return TRUE;
}

BOOL ProcessorSse2Aroo::wndProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam, void *editp, FILTER *fp)
{
	if(message == WM_FILTER_CHANGE_ACTIVE && !fp->exfunc->is_filter_active(fp)) {
		mem.deallocate();
	}
	return FALSE;
}
